from yggdrasil.serialize.tests import test_AsciiTableSerialize as parent


class TestPandasSerialize(parent.TestAsciiTableSerialize):
    r"""Test class for TestPandasSerialize class."""

    _cls = 'PandasSerialize'

    def test_apply_field_names_errors(self):
        r"""Test errors raised by apply_field_names."""
        self.assert_raises(RuntimeError, self.instance.apply_field_names,
                           self.testing_options['objects'][0],
                           field_names=['x', 'y'])
        names = self.testing_options['objects'][0].columns.tolist()
        names[0] = 'invalid'
        self.assert_raises(RuntimeError, self.instance.apply_field_names,
                           self.testing_options['objects'][0],
                           field_names=names)

    def test_func_serialize_errors(self):
        r"""Test errors raised by func_serialize."""
        self.assert_raises(TypeError, self.instance.func_serialize, None)

    def test_deserialize_no_header(self):
        r"""Test deserialization of frame output without a header."""
        if self.testing_option_kws.get('no_header', False):
            return
        kws = self.instance.get_testing_options(no_header=True)
        kws['kwargs'].pop('no_header', None)
        no_head_inst = self.import_cls(**kws['kwargs'])
        x = no_head_inst.serialize(kws['objects'][0])
        y = self.instance.deserialize(x)[0]
        self.assert_result_equal(y, self.testing_options['objects'][0])


class TestPandasSerializeNoHeader(TestPandasSerialize):
    r"""Test class for PandasSerialize class when no header specified."""

    testing_option_kws = {'no_header': True}

    
class TestPandasSerializeBytes(TestPandasSerialize):
    r"""Test class for PandasSerialize class when strings are bytes."""

    testing_option_kws = {'table_string_type': 'bytes'}



from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent
from yggdrasil.metaschema.properties.tests.test_ArgsMetaschemaProperty import (
    ValidArgsClass4, InvalidArgsClass)


class TestInstanceMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for InstanceMetaschemaType class."""

    _mod = 'InstanceMetaschemaType'
    _cls = 'InstanceMetaschemaType'

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestMetaschemaType.after_class_creation(cls)
        cls._typedef.update({'class': ValidArgsClass4,
                             'args': ValidArgsClass4.valid_args,
                             'kwargs': ValidArgsClass4.valid_kwargs})
        value = ValidArgsClass4(*ValidArgsClass4.test_args,
                                **ValidArgsClass4.test_kwargs)
        cls._valid_encoded = [dict(cls._typedef,
                                   type=cls.get_import_cls().name)]
        cls._valid_decoded = [value]
        cls._invalid_encoded = [{}]
        cls._invalid_decoded = [int(1), 'hello']
        cls._invalid_validate += [InvalidArgsClass()]
        cls._compatible_objects = [(value, value, None)]



import os
import pandas as pd
from yggdrasil import units, tools, multitasking
from yggdrasil.drivers.DSLModelDriver import DSLModelDriver


_default_agg = 'mean'
_default_interp = 'index'


class TimeSyncModelDriver(DSLModelDriver):
    r"""Class for synchronizing states for timesteps between two models.

    Args:
        synonyms (dict, optional): Mapping from model names to mappings
            from base variables names to information about one or
            more alternate variable names used by the named model
            that should be converted to the base variable. Values for
            providing information about alternate variables can either
            be strings (implies equivalence with the base variable in
            everything but name and units) or mappings with the keys:

              alt (str, list): Name of one or more variables used by
                 the model that should be used to calculate the named
                 base variable.
              alt2base (function): Callable object that takes the
                 alternate variables named by the 'alt' property as
                 input and returns the base variable.
              base2alt (function): Callable object that takes the base
                 variable as input and returns the alternate variables
                 named by the 'alt' property.

            Defaults to an empty dictionary.
        aggregation (str, dict, optional): Method(s) that should be used
            to aggregate synonymous variables across models. This can
            be a single method that should be used for all synonymous
            variables or a dictionary mapping between synonymous variables
            and the method that should be used by that variable. If a
            variable is not present in the dictionary or a values is
            not provided, 'mean' will be used. See the documentation
            for pandas.DataFrame.aggregate for available options.
        interpolation (str, dict, optional): Method(s) or keyword
            arguments that should be used to interpolate missing
            timesteps. This can be a single method or a dictionary
            mapping between model name and the interpolation methods
            (or keyword arguments) that should be used for variables
            from that model. Defaults to 'index'. See the documentation
            for pandas.DataFrame.interpolate for available options.
        additional_variables (dict, optional): Mapping from model
            name to a list of variables from other models that are
            not provided by the model, but should still be returned
            to the model. Defaults to empty dictionary.

    """

    _schema_subtype_description = ('Model is dedicated to synchronizing'
                                   'timesteps between other models.')
    _schema_properties = {
        'synonyms': {'type': 'object',
                     'additionalProperties': {
                         'type': 'object',
                         'additionalProperties': {'anyOf': [
                             {'type': 'string'},
                             {'type': 'object',
                              'required': ['alt', 'alt2base', 'base2alt'],
                              'properties': {
                                  'alt': {'anyOf': [
                                      {'type': 'string'},
                                      {'type': 'array',
                                       'items': {'type': 'string'}}]},
                                  'alt2base': {'type': 'function'},
                                  'base2alt': {'type': 'function'}}}]}},
                     'default': {}},
        'interpolation': {
            'anyOf': [{'type': 'string'},
                      {'type': 'object',
                       'additionalProperties': {'oneOf': [
                           {'type': 'string'},
                           {'type': 'object',
                            'required': ['method'],
                            'properties': {
                                'method': {'type': 'string'}}}]}},
                      {'type': 'object',
                       'required': ['method'],
                       'properties': {
                           'method': {'type': 'string'}}}],
            'default': _default_interp},
        'aggregation': {
            'anyOf': [{'type': 'function'},
                      {'type': 'string'},
                      {'type': 'object',
                       'additionalProperties': {
                           'anyOf': [{'type': 'function'},
                                     {'type': 'string'}]}}],
            'default': _default_agg},
        'additional_variables': {
            'type': 'object',
            'additionalProperties': {'type': 'array',
                                     'items': {'type': 'string'}},
            'default': {}}}
    language = 'timesync'

    def __init__(self, name, *args, **kwargs):
        super(TimeSyncModelDriver, self).__init__(name, *args, **kwargs)
        # Ensure that options are uniform in their format and check
        # that they are valid
        for k, v in self.synonyms.items():
            for s0, x in list(v.items()):
                if isinstance(x, str):
                    x = {'alt': x, 'alt2base': None, 'base2alt': None}
                if not isinstance(x['alt'], list):
                    x['alt'] = [x['alt']]
                if ((((x['alt2base'] is None) or (x['base2alt'] is None))
                     and (len(x['alt']) > 1))):  # pragma: debug
                    raise RuntimeError(
                        ('Cannot convert from multiple alternate '
                         'variables (%s) to single base variable (%s) '
                         'without transformation functions.')
                        % (x['alt'], s0))
                v[s0] = x
        if isinstance(self.interpolation, str):
            self.interpolation = {'method': self.interpolation}
        if 'method' not in self.interpolation:
            for k, v in list(self.interpolation.items()):
                if isinstance(v, str):
                    v = {'method': v}
                    self.interpolation[k] = v

    def parse_arguments(self, args, **kwargs):
        r"""Sort model arguments to determine which one is the executable
        and which ones are arguments.

        Args:
            args (list): List of arguments provided.
            **kwargs: Additional keyword arguments are ignored.

        """
        assert(isinstance(args, list) and (len(args) == 0))
        self.model_file = 'dummy'
        
    @property
    def model_wrapper_args(self):
        r"""tuple: Positional arguments for the model wrapper."""
        return (self.name, self.synonyms, self.interpolation,
                self.aggregation, self.additional_variables)

    @classmethod
    def model_wrapper(cls, name, synonyms, interpolation,
                      aggregation, additional_variables, env=None):
        r"""Model wrapper."""
        from yggdrasil.languages.Python.YggInterface import YggRpcServer
        if env is not None:
            os.environ.update(env)
        rpc = YggRpcServer(name)
        threads = {}
        times = []
        tables = {}
        table_units = {'base': {}}
        table_lock = multitasking.RLock()
        default_agg = _default_agg
        if not isinstance(aggregation, dict):
            default_agg = aggregation
            aggregation = {}
        while True:
            # Check for errors on response threads
            for v in threads.values():
                if v.check_flag_attr('error_flag'):  # pragma: debug
                    for v in threads.values():
                        if v.is_alive():
                            v.terminate()
                    raise Exception("Error on response thread.")
            # Receive values from client models
            flag, values, request_id = rpc.recv_from(timeout=1.0)
            if not flag:
                print("timesync server: End of input.")
                break
            if len(values) == 0:
                rpc.sleep()
                continue
            t, state = values[:]
            t_pd = units.convert_to_pandas_timedelta(t)
            client_model = rpc.ocomm[request_id].client_model
            # Update record
            with table_lock:
                if client_model not in tables:
                    tables[client_model] = pd.DataFrame({'time': times})
                # Update units & aggregation methods
                if client_model not in table_units:
                    # NOTE: this assumes that units will not change
                    # between timesteps for a single model. Is there a
                    # case where this might not be true?
                    table_units[client_model] = {
                        k: units.get_units(v) for k, v in state.items()}
                    table_units[client_model]['time'] = units.get_units(t)
                    alt_vars = []
                    for k, v in synonyms.get(client_model, {}).items():
                        alt_vars += v['alt']
                        if v['alt2base'] is not None:
                            table_units[client_model][k] = units.get_units(
                                v['alt2base'](*[state[a] for a in v['alt']]))
                        else:
                            table_units[client_model][k] = table_units[
                                client_model][v['alt'][0]]
                    for k, v in table_units[client_model].items():
                        table_units['base'].setdefault(k, v)
                    for k in list(set(state.keys()) - set(alt_vars)):
                        aggregation.setdefault(k, default_agg)
                # Update the state
                if t_pd not in times:
                    times.append(t_pd)
                for model, table in tables.items():
                    new_data = {'time': [t_pd]}
                    if model == client_model:
                        new_data.update({k: [units.get_data(v)]
                                         for k, v in state.items()})
                    new_data = pd.DataFrame(new_data)
                    idx = table['time'].isin([t_pd])
                    if not idx.any():
                        table = table.append(new_data, sort=False)
                    elif model == client_model:
                        table = table.drop(table.index[idx])
                        table = table.append(new_data, sort=False)
                    tables[model] = table.sort_values('time')
            # Assign thread to handle checking when data is filled in
            internal_variables = list(state.keys())
            external_variables = additional_variables.get(client_model, [])
            threads[request_id] = multitasking.YggTaskLoop(
                target=cls.response_loop,
                args=(client_model, request_id, rpc, t_pd,
                      internal_variables, external_variables,
                      tables, table_units, table_lock,
                      synonyms, interpolation, aggregation))
            threads[request_id].start()
        # Cleanup threads (only called if there is an error since the
        # loop will only be broken when all of the clients have signed
        # off, implying that all requests have been responded to).
        for v in threads.values():
            if v.is_alive():  # pragma: debug
                v.wait(0.5)
        for v in threads.values():
            if v.is_alive():  # pragma: debug
                v.terminate()

    @classmethod
    def check_for_data(cls, time, tables, table_lock, open_clients):
        r"""Check for a time in the tables to determine if there is
        sufficient data available to calculate the state.

        Args:
            time (pandas.Timedelta): Time that state is requested at.
            tables (dict): Mapping from model name to pandas DataFrames
                containing variables supplied by the model.
            table_lock (RLock): Thread-safe lock for accessing table.
            open_clients (list): Clients that are still open.

        Returns:
            bool: True if there is sufficient data, False otherwise.

        """
        with table_lock:
            for k, v in tables.items():
                if (k in open_clients) and (time > max(v.dropna()['time'])):
                    return False
        return True

    @classmethod
    def response_loop(cls, client_model, request_id, rpc, time,
                      internal_variables, external_variables,
                      tables, table_units, table_lock,
                      synonyms, interpolation, aggregation):
        r"""Check for available data and send response if it is
        available.

        Args:
            client_model (str): Name of model that made the request.
            request_id (str): ID associated with request that should
                be responded to.
            rpc (ServerComm): Server RPC comm that should be used to
                reply to the request when the data is available.
            time (pandas.Timedelta): Time to get variables at.
            internal_variables (list): Variables that model is requesting
                that it also calculates.
            external_variables (list): Variables that model is requesting
                that will be provided by other models.
            tables (dict): Mapping from model name to pandas DataFrames
                containing variables supplied by the model.
            table_units (dict): Mapping from model name to dictionaries
                mapping from variable names to units.
            table_lock (RLock): Thread-safe lock for accessing table.
            synonyms (dict): Dictionary mapping from base variables to
                alternate variables and mapping functions used to convert
                between the variables. Defaults to empty dict and no
                conversions are performed.
            interpolation (dict): Mapping from model name to the
                interpolation kwargs that should be used. Defaults to
                empty dictionary.
            aggregation (dict): Mapping from variable name to the
                aggregation method that should be used. Defaults to
                empty dictionary.

        """
        if not (rpc.all_clients_connected
                and cls.check_for_data(time, tables, table_lock,
                                       rpc.open_clients)):
            # Don't start sampling until all clients have connected
            # and there is data available for the requested timestep
            tools.sleep(1.0)
            return
        tot = cls.merge(tables, table_units, table_lock, rpc.open_clients,
                        synonyms, interpolation, aggregation)
        # Update external units
        for k in external_variables:
            if k not in table_units[client_model]:
                table_units[client_model][k] = table_units['base'][k]
        # Check if data is available at the desired timestep?
        # Convert units
        for k in tot.columns:
            funits = units.get_conversion_function(table_units['base'][k],
                                                   table_units[client_model][k])
            tot[k] = tot[k].apply(funits)
        # Transform back to variables expected by the model
        for kbase, alt in synonyms.get(client_model, {}).items():
            if alt['base2alt'] is not None:
                alt_vars = alt['base2alt'](tot[kbase])
                if isinstance(alt_vars, (tuple, list)):
                    assert(len(alt_vars) == len(alt['alt']))
                    for k, v in zip(alt['alt'], alt_vars):
                        tot[k] = v
                else:
                    assert(len(alt['alt']) == 1)
                    tot[alt['alt'][0]] = alt_vars
            else:
                tot[alt['alt'][0]] = tot[kbase]
        # Get state
        state = {}
        for v in internal_variables + external_variables:
            v_res = tot.loc[time, v]
            state[v] = units.add_units(v_res, table_units[client_model][v])
        time_u = units.convert_to(units.convert_from_pandas_timedelta(time),
                                  table_units[client_model]['time'])
        flag = rpc.send_to(request_id, state)
        if not flag:  # pragma: debug
            raise RuntimeError(("Failed to send response to "
                                "request %s for time %s from "
                                "model %s.")
                               % (request_id, time_u, client_model))
        raise multitasking.BreakLoopException
    
    @classmethod
    def merge(cls, tables, table_units, table_lock, open_clients,
              synonyms, interpolation, aggregation):
        r"""Merge tables from models to get data.

        Args:
            tables (dict): Mapping from model name to pandas DataFrames
                containing variables supplied by the model.
            table_units (dict): Mapping from model name to dictionaries
                mapping from variable names to units.
            table_lock (RLock): Thread-safe lock for accessing table.
            open_clients (list): Clients that are still open.
            synonyms (dict): Dictionary mapping from base variables to
                alternate variables and mapping functions used to convert
                between the variables. Defaults to empty dict and no
                conversions are performed.
            interpolation (dict): Mapping from model name to the
                interpolation kwargs that should be used. Defaults to
                empty dictionary.
            aggregation (dict): Mapping from variable name to the
                aggregation method that should be used. Defaults to
                empty dictionary.


        """
        # Adjust input arguments
        interp_default = {'method': _default_interp}
        if 'method' in interpolation:
            interp_default = interpolation
            interpolation = {}
        # Interpolate
        table_temp = {}
        with table_lock:
            for k, v in tables.items():
                kws = interpolation.get(k, interp_default).copy()
                if k not in open_clients:
                    # Ensure that clients that have signed of are
                    # extrapolated, otherwise they would never produce
                    # valid data
                    kws['limit_area'] = None
                if 'order' in kws:
                    kws['order'] = min(v.dropna().shape[0] - 1,
                                       kws['order'])
                    if kws['order'] == 0:
                        kws.pop('order')
                        kws.update(interp_default)
                v = v.set_index('time')
                # Cannot interpolate on pandas timedelta as of pandas 1.0.1
                ind = v.index
                v.index = v.index.total_seconds()
                table_temp[k] = v.interpolate(**kws)
                table_temp[k].index = ind
        # Rename + transformation
        for model, v in table_temp.items():
            drop = []
            for kbase, alt in synonyms.get(model, {}).items():
                if alt['alt2base'] is not None:
                    args = [v[k] for k in alt['alt']]
                    v[kbase] = alt['alt2base'](*args)
                else:
                    v[kbase] = v[alt['alt'][0]]
                drop += alt['alt']
            for k in drop:
                v = v.drop(k, axis=1)
            # Units
            for k in v.columns:
                funits = units.get_conversion_function(table_units[model][k],
                                                       table_units['base'][k])
                v[k] = v[k].apply(funits)
            table_temp[model] = v
        # Append
        out = pd.DataFrame()
        for k, v in table_temp.items():
            out = out.append(v, sort=False)
        # Groupby + aggregate
        out = out.groupby('time').agg(aggregation)
        return out



import numpy as np
from yggdrasil import units
from yggdrasil.tools import safe_eval
from yggdrasil.communication.transforms.TransformBase import TransformBase


class StatementTransform(TransformBase):
    r"""Class for transforming messages based on a provided statement using Python syntax.

    Args:
        statement (str): Python statement in terms of the message as represented by
            the string "%x%" that should evaluate to the transformed message.
            The statement should only use a limited set of builtins and the math
            library (See yggdrasil.tools.safe_eval). If more complex relationships
            are required, use the FunctionTransform class.

    Attributes:
        statement (str): Python statement that will be evaluated to transform
            messages.

    """
    _transformtype = 'statement'
    _schema_required = ['statement']
    _schema_properties = {'statement': {'type': 'string'}}

    def __init__(self, *args, **kwargs):
        super(StatementTransform, self).__init__(*args, **kwargs)
        self.statement = self.statement.replace('%x%', 'x')

    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            bool: True if the message will pass through the transform, False otherwise.

        """
        return safe_eval(self.statement, x=x)

    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        out = [{'kwargs': {'statement': '%x%**3'},
                'in/out': [(1, 1), (2, 8)]},
               {'kwargs': {'statement': '%x% * array([1, 1, 1])'},
                'in/out': [(1, np.ones(3, int)), (2, 2 * np.ones(3, int))]},
               {'kwargs': {'statement': '%x% * '
                           + repr(units.add_units(1, 'cm'))},
                'in/out': [(1, units.add_units(1, 'cm')),
                           (2, units.add_units(2, 'cm'))]}]
        return out



from yggdrasil import multitasking
from yggdrasil.components import import_component
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver


class DSLModelDriver(InterpretedModelDriver):  # pragma: no cover
    r"""Class for running domain specific lanugage models."""

    is_dsl = True
    base_languages = ['python']  # Defaults to Python but can be modified
    executable_type = 'dsl'
    function_param = None
    full_language = False

    @classmethod
    def is_language_installed(self):
        r"""Determine if this model driver is installed on the current
        machine.

        Returns:
            bool: Truth of if this model driver can be run on the current
                machine.

        """
        # Dependent only on base languages
        return True

    @classmethod
    def is_library_installed(cls, lib, **kwargs):
        r"""Determine if a dependency is installed.

        Args:
            lib (str): Name of the library that should be checked.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            bool: True if the library is installed, False otherwise.

        """
        drv = import_component('model', cls.base_languages[0])
        return drv.is_library_installed(lib, **kwargs)
    
    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.run_executable.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        return None
    
    @classmethod
    def model_wrapper(cls, *args, **kwargs):  # pragma: no cover
        r"""Model wrapper."""
        raise NotImplementedError

    @property
    def model_wrapper_args(self):  # pragma: no cover
        r"""tuple: Positional arguments for the model wrapper."""
        return ()

    @property
    def model_wrapper_kwargs(self):
        r"""dict: Keyword arguments for the model wrapper."""
        return {'env': self.set_env()}

    def queue_recv(self):
        r"""Receive a message from the model process."""
        while not (self.model_process.pipe[0].poll()
                   or self.model_process.pipe[0].closed
                   or self.model_process.pipe[1].closed
                   or (not self.model_process.is_alive())
                   or self.queue_thread.was_break):
            self.sleep()
        if not self.model_process.pipe[0].poll():
            raise RuntimeError("No more messages from model process.")
        out = self.model_process.pipe[0].recv()
        if isinstance(out, str):
            out = out.encode('utf-8')
        return out
        
    def queue_close(self):
        r"""Close the queue for messages from the model process."""
        self.model_process.pipe[0].close()
        self.model_process.pipe[1].close()
        
    def run_model(self, return_process=True, **kwargs):
        r"""Run the model. Unless overridden, the model will be run using
        run_executable.

        Args:
            return_process (bool, optional): If True, the process running
                the model is returned. If False, the process will block until
                the model finishes running. Defaults to True.
            **kwargs: Keyword arguments are passed to run_executable.

        """
        args = self.model_wrapper_args
        kwargs = self.model_wrapper_kwargs
        self.debug('Working directory: %s', self.working_dir)
        self.debug('Model file: %s', self.model_file)
        self.debug('Environment Variables:\n%s',
                   self.pprint(kwargs['env'], block_indent=1))
        p = multitasking.YggTask(task_method='process', with_pipe=True,
                                 target=self.model_wrapper,
                                 args=args, kwargs=kwargs)
        p.start()
        if return_process:
            return p
        p.join()
        if p.returncode != 0:
            raise RuntimeError("Model failed.")
        return ''

    def after_loop(self):
        r"""Actions to perform after run_loop has finished. Mainly checking
        if there was an error and then handling it."""
        super(DSLModelDriver, self).after_loop()
        self.info("returncode = %d", self.model_process.returncode)
        if self.model_process.returncode != 0:
            self.error("Error on model process.")



from yggdrasil.communication.tests import test_CommBase


class TestBufferComm(test_CommBase.TestCommBase):
    r"""Tests for BufferComm communication class."""

    comm = 'BufferComm'

    def test_error_name(self):
        r"""Test error on missing address."""
        # Skipped because the name is not used to intialize
        # the address for a buffer
        pass



import numpy as np
import pandas
from yggdrasil.communication.tests import test_AsciiTableComm as parent


class TestPandasFileComm(parent.TestAsciiTableComm):
    r"""Test for PandasFileComm communication class."""

    comm = 'PandasFileComm'


class TestPandasFileComm_nonames(TestPandasFileComm):
    r"""Test for PandasFileComm communication class without field names sent."""

    testing_option_kws = {'no_names': True}


class TestPandasFileComm_single(TestPandasFileComm):
    r"""Test for PandasFileComm communication class with field names sent."""

    def get_options(self):
        r"""Get testing options."""
        nele = 5
        dtype = np.dtype(dict(formats=['float'], names=['f0']))
        arr1 = np.zeros((nele, ), dtype)
        arr2 = np.ones((nele, ), dtype)
        pd1 = pandas.DataFrame(arr1)
        pd2 = pandas.DataFrame(arr2)
        pdcat = pandas.DataFrame(np.hstack([arr1, arr2]))
        out = {'kwargs': {},
               'contents': (b'f0\n' + nele * b'0.0\n' + nele * b'1.0\n'),
               'send': [pd1, pd2],
               'recv': [pdcat],
               'recv_partial': [[pd1], [pdcat]],
               'dict': {'f0': arr1['f0']},
               'objects': [pd1, pd2]}
        out['msg'] = out['send'][0]
        out['msg_array'] = arr1
        return out

    def test_send_dict_default(self):
        r"""Test automated conversion of dictionary to pandas data frame."""
        self.do_send_recv(msg_send=self.testing_options['dict'],
                          msg_recv=self.testing_options['msg'])



from yggdrasil.tests import generate_component_tests
from yggdrasil.communication.filters.tests.test_FilterBase import TestFilterBase


generate_component_tests('filter', TestFilterBase,
                         globals(), __file__, class_attr='filter')


__all__ = []



import os
import copy
import glob
from yggdrasil import components
from yggdrasil.drivers.CompiledModelDriver import (
    CompiledModelDriver, CompilerBase)


class BuildToolBase(CompilerBase):
    r"""Base class for build tools."""

    is_build_tool = True
    build_language = None

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        CompilerBase.before_registration(cls)
        if cls.build_language is None:
            cls.build_language = cls.toolname
        
    @classmethod
    def get_default_target_language(cls):
        r"""Determine the default target language for the build tool.
        Unless otherwise specified, this will be the first language in
        the 'languages' attribute that is not the build tool.

        Returns:
            str: Name of the default target language.

        """
        build_language = cls.build_language
        for x in cls.languages:
            if x != build_language:
                return x
        raise ValueError("Could not determine a default target language "
                         "for build tool '%s'" % cls.toolname)  # pragma: debug

    @classmethod
    def set_env(cls, language=None, language_driver=None, **kwargs):
        r"""Get environment variables that should be set for the model process.

        Args:
            language (str, optional): Language that is being compiled. Defaults
                to the first language in cls.languages that isn't toolname.
            language_driver (ModelDriver, optional): Driver for language that
                should be used. Defaults to None and will be imported based
                on language.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(BuildToolBase, cls).set_env(**kwargs)
        if language_driver is None:
            if language is None:
                language = cls.get_default_target_language()
            language_driver = components.import_component('model', language)
        compiler = language_driver.get_tool('compiler')
        out = compiler.set_env(existing=out)
        return out


class BuildModelDriver(CompiledModelDriver):
    r"""Class for running build file compiled drivers.

    Args:
        name (str): Driver name.
        args (str, list): Executable that should be created (make target) and
            any arguments for the executable.
        target (str, optional): Make target that should be built to create the
            model executable. Defaults to None.
        target_language (str, optional): Language that the target is written in.
            Defaults to None and will be set based on the source files provided.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        buildfile (str): Path to file containing build instructions.
        builddir (str): Path to directory where build products will be saved.
        sourcedir (str): Path to directory where source files are located.
        compiledir (str): Path to directory where compilation call should be
            made from. Defaults to working_dir.
        target (str): Name of executable that should be created and called.
        target_language (str): Language that the target is written in.
        target_language_driver (ModelDriver): Language driver for the target
            language.

    Class Attributes:
        built_where_called (bool): If True, it is assumed that compilation
            output will be saved in the same directory from which the
            compilation command is issued.

    Raises:
        RuntimeError: If neither the IPC or ZMQ C libraries are available.

    """

    base_languages = ['c', 'c++']
    built_where_called = False
    sourcedir_as_sourcefile = False
    full_language = False
    is_build_tool = True

    def __init__(self, *args, **kwargs):
        self.target_language_driver = None
        super(BuildModelDriver, self).__init__(*args, **kwargs)

    @staticmethod
    def after_registration(cls, **kwargs):
        r"""Operations that should be performed to modify class attributes after
        registration. For compiled languages this includes selecting the
        default compiler. The order of precedence is the config file 'compiler'
        option for the language, followed by the environment variable set by
        _compiler_env, followed by the existing class attribute.
        """
        for k in ['linker', 'archiver']:
            if k in cls._config_keys:
                cls._config_keys.remove(k)
        if getattr(cls, 'default_compiler', None) is None:
            cls.default_compiler = cls.language
        CompiledModelDriver.after_registration(cls, **kwargs)
        
    def parse_arguments(self, args, **kwargs):
        r"""Sort arguments based on their syntax to determine if an argument
        is a source file, compilation flag, or runtime option/flag that should
        be passed to the model executable.

        Args:
            args (list): List of arguments provided.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        """
        # Set builddir before passing to parent class so that builddir is used
        # to normalize the model file path rather than the working directory
        # which may be different.
        default_attr = [('target_language_driver', None),
                        ('buildfile', None),
                        ('builddir', None),
                        ('sourcedir', None),
                        ('compile_working_dir', None),
                        ('builddir_base', '.'),
                        ('buildfile_base', None)]
        for k, v in default_attr:
            if not hasattr(self, k):
                setattr(self, k, v)
        # Directory that compilation should be called from
        if ((isinstance(self.compile_working_dir, str)
             and not os.path.isabs(self.compile_working_dir))):
            self.compile_working_dir = os.path.realpath(
                os.path.join(self.working_dir, self.compile_working_dir))
        # Source directory
        if self.sourcedir is None:
            self.sourcedir = os.path.dirname(args[0])
        if not os.path.isabs(self.sourcedir):
            self.sourcedir = os.path.normpath(
                os.path.realpath(os.path.join(self.working_dir,
                                              self.sourcedir)))
        # Build file
        if self.buildfile is None:
            self.buildfile = self.buildfile_base
        if not os.path.isabs(self.buildfile):
            for x in set([self.working_dir, self.sourcedir,
                          self.builddir, self.compile_working_dir]):
                if x is not None:
                    y = os.path.normpath(os.path.join(x, self.buildfile))
                    if os.path.isfile(y):
                        self.buildfile = y
                        break
        # Compilation directory
        if self.compile_working_dir is None:
            self.compile_working_dir = os.path.dirname(self.buildfile)
        if not os.path.isabs(self.compile_working_dir):
            self.compile_working_dir = os.path.realpath(
                os.path.join(self.working_dir, self.compile_working_dir))
        # Build directory
        if self.builddir is None:
            if self.built_where_called:
                self.builddir = self.compile_working_dir
            else:
                self.builddir = os.path.join(self.sourcedir, self.builddir_base)
        if not os.path.isabs(self.builddir):
            self.builddir = os.path.realpath(os.path.join(self.working_dir,
                                                          self.builddir))
        kwargs.setdefault('default_model_dir', self.builddir)
        if self.sourcedir_as_sourcefile:
            self.source_files = [self.sourcedir]
        super(BuildModelDriver, self).parse_arguments(args, **kwargs)

    def set_target_language(self):
        r"""Set the language of the target being compiled (usually the same
        as the language associated with this driver.

        Returns:
            str: Name of language.

        """
        source_dir = self.sourcedir
        if self.target_language is None:
            # Get file form cmakelists files
            try_list = sorted(list(glob.glob(os.path.join(source_dir, '*'))))
            early_exit = False
            if self.model_src is not None:
                try_list = [self.model_src, try_list]
                early_exit = True
            languages = copy.deepcopy(self.get_tool_instance('compiler').languages)
            languages.remove(self.language)
            try:
                self.target_language = self.get_language_for_source(
                    try_list, early_exit=early_exit, languages=languages)
            except ValueError:
                pass
            # Try to compile C as C++
            # if self.target_language == 'c':
            #     self.target_language = 'c++'
        if (((self.target_language_driver is None)
             and (self.target_language is not None))):
            self.target_language_driver = components.import_component(
                'model', self.target_language)
        return self.target_language
        
    @classmethod
    def is_source_file(cls, fname):
        r"""Determine if the provided file name points to a source files for
        the associated programming language by checking the extension.

        Args:
            fname (str): Path to file.

        Returns:
            bool: True if the provided file is a source file, False otherwise.

        """
        compiler = cls.get_tool('compiler')
        for lang in compiler.languages:
            if lang == cls.language:
                continue
            drv = components.import_component('model', lang)
            if drv.is_source_file(fname):
                return True
        return False
        
    def compile_model(self, **kwargs):
        r"""Compile model executable(s).

        Args:
            **kwargs: Keyword arguments are passed on to the parent class's
                method.

        """
        if self.target_language_driver is not None:
            self.target_language_driver.compile_dependencies()
        kwargs['working_dir'] = self.compile_working_dir
        return super(BuildModelDriver, self).compile_model(**kwargs)
        
    def cleanup(self):
        r"""Remove compiled executable."""
        if (self.model_file is not None) and os.path.isfile(self.model_file):
            self.compile_model(target='clean')
        super(BuildModelDriver, self).cleanup()

    def set_env(self, **kwargs):
        r"""Get environment variables that should be set for the model process.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        if kwargs.get('for_compile', False) and (self.target_language is not None):
            kwargs.setdefault('compile_kwargs', {})
            kwargs['compile_kwargs']['language'] = self.target_language
            kwargs['compile_kwargs']['language_driver'] = self.target_language_driver
        out = super(BuildModelDriver, self).set_env(**kwargs)
        if not kwargs.get('for_compile', False):
            if hasattr(self.target_language_driver, 'update_ld_library_path'):
                self.target_language_driver.update_ld_library_path(out)
        return out



import copy
import numpy as np
from yggdrasil.communication.transforms.TransformBase import TransformBase


class MapFieldsTransform(TransformBase):
    r"""Class for mapping a subset of the original fields in an object to
    a different set of fields. Fields that don't exist in the map are preserved
    unchanged.

    Args:
        map (dict): A mapping from original field name to new field names.

    """
    _transformtype = 'map_fields'
    _schema_required = ['map']
    _schema_properties = {'map': {'type': 'object',
                                  'additionalProperties': {'type': 'string'}}}

    def transform_datatype(self, datatype):
        r"""Determine the datatype that will result from applying the transform
        to the supplied datatype.

        Args:
            datatype (dict): Datatype to transform.

        Returns:
            dict: Transformed datatype.

        """
        if (((datatype.get('type', None) == 'array')
             and isinstance(datatype.get('items', None), list))):
            datatype = copy.deepcopy(datatype)
            for i, x in enumerate(datatype['items']):
                if x.get('title', 'f%d' % i) in self.map:
                    x['title'] = self.map[x['title']]
        elif datatype.get('type', None) == 'object':
            datatype = copy.deepcopy(datatype)
            for kold, knew in self.map.items():
                datatype['properties'][knew] = datatype['properties'].pop(kold)
        return datatype
    
    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        out = x
        if isinstance(x, dict):
            if not no_copy:
                out = copy.deepcopy(x)
            for kold, knew in self.map.items():
                out[knew] = out.pop(kold)
        elif isinstance(x, (list, tuple)):
            pass
        elif isinstance(x, np.ndarray):
            if not no_copy:
                out = copy.deepcopy(x)
            new_names = list(x.dtype.names)
            for kold, knew in self.map.items():
                new_names[new_names.index(kold)] = knew
            out.dtype.names = new_names
        else:
            raise TypeError("Cannot map fields from object of type '%s'" % type(x))
        return out
    
    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        return [{'kwargs': {'map': {'a': 'aa', 'c': 'cc'}},
                 'in/out': [(dict(zip('abc', range(3))),
                             {'aa': 0, 'b': 1, 'cc': 2})],
                 'in/out_t': [
                     ({'type': 'object',
                       'properties': {
                           x: {'type': 'int'}
                           for x in 'abc'}},
                      {'type': 'object',
                       'properties': {
                           x: {'type': 'int'}
                           for x in ['aa', 'b', 'cc']}})]},
                {'kwargs': {'map': {'a': 'aa', 'c': 'cc'}},
                 'in/out': [([0, 1, 2], [0, 1, 2])],
                 'in/out_t': [
                     ({'type': 'array',
                       'items': [{'type': 'int', 'title': x}
                                 for x in 'abc']},
                      {'type': 'array',
                       'items': [{'type': 'int', 'title': x}
                                 for x in ['aa', 'b', 'cc']]})]},
                {'kwargs': {'map': {'a': 'aa', 'c': 'cc'}},
                 'in/out': [(np.zeros(3, np.dtype({'names': ['a', 'b', 'c'],
                                                   'formats': ['i4', 'i4', 'i4']})),
                             np.zeros(3, np.dtype({'names': ['aa', 'b', 'cc'],
                                                   'formats': ['i4', 'i4', 'i4']})))]},
                {'kwargs': {'map': {'a': 'aa', 'c': 'cc'}},
                 'in/out': [(None, TypeError)]}]



from yggdrasil.metaschema.datatypes import MetaschemaTypeError
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent
from yggdrasil.metaschema.datatypes.MultiMetaschemaType import (
    create_multitype_class)


class TestMultiMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for MultiMetaschemaType class."""

    _mod = 'MultiMetaschemaType'
    _cls = 'MultiMetaschemaType'
    _types = ['object', 'int']
    
    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestMetaschemaType.after_class_creation(cls)
        cls._value = {'a': int(1), 'b': float(1)}
        cls._valid_encoded = [dict(cls._typedef,
                                   type=cls._types)]
        cls._valid_decoded = [cls._value, int(1)]
        cls._invalid_decoded = ['hello']
        cls._compatible_objects = [(cls._value, cls._value, None)]

    @property
    def import_cls(self):
        r"""Import the tested class from its module"""
        return create_multitype_class(self._types)
        
    @property
    def typedef(self):
        r"""dict: Type definition."""
        out = super(TestMultiMetaschemaType, self).typedef
        out['type'] = self._types
        return out

    def test_type_mismatch_error(self):
        r"""Test that error is raised when there is a type mismatch."""
        self.assert_raises(MetaschemaTypeError, self.import_cls,
                           type=['invalid'])



import pandas
import numpy as np
from yggdrasil.communication.transforms.ArrayTransform import ArrayTransform
from yggdrasil.serialize import numpy2pandas


class PandasTransform(ArrayTransform):
    r"""Class for consolidating values into a Pandas data frame."""
    _transformtype = 'pandas'

    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        if isinstance(x, pandas.DataFrame):
            out = x
        else:
            out = super(PandasTransform, self).evaluate_transform(
                x, no_copy=no_copy)
            out = numpy2pandas(out)
        return out
    
    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        length = 5
        t = {'type': 'array',
             'items': [
                 {'type': '1darray', 'subtype': 'bytes',
                  'precision': 40, 'length': length},
                 {'type': '1darray', 'subtype': 'int',
                  'precision': 64, 'length': length},
                 {'type': '1darray', 'subtype': 'float',
                  'precision': 64, 'length': length},
                 {'type': '1darray', 'subtype': 'complex',
                  'precision': 128, 'length': length}]}
        dtype = np.dtype([(n, f) for n, f in zip(
            ['f0', 'f1', 'f2', 'f3'], ['S5', 'i8', 'f8', 'c16'])])
        x = np.zeros(length, dtype=dtype)
        x[dtype.names[0]][0] = b'hello'
        y = [x[n] for n in dtype.names]
        x = numpy2pandas(x)
        return [{'kwargs': {'original_datatype': t},
                 'in/out': [(y, x)],
                 'in/out_t': [(t, t)]},
                {'kwargs': {'original_datatype': t},
                 'in/out': [(x, x)],
                 'in/out_t': [(t, t)]},
                {'kwargs': {'original_datatype': t},
                 'in/out': [(None, TypeError)]},
                {'kwargs': {},
                 'in/out': [([0, 1, 2], AssertionError)]}]



import re
import numpy as np
import pandas as pd
import unyt
from yggdrasil import tools
_unit_quantity = unyt.array.unyt_quantity
_unit_array = unyt.array.unyt_array
_ureg_unyt = None


def get_ureg():
    r"""Get the unit registry."""
    global _ureg_unyt
    if _ureg_unyt is None:
        _ureg_unyt = unyt.UnitRegistry('mks')
        _ureg_unyt.add("ac", 4046.86, dimensions=unyt.dimensions.area,
                       tex_repr=r"\rm{ac}", offset=0.0, prefixable=False)
        _ureg_unyt.add("a", 100.0, dimensions=unyt.dimensions.area,
                       tex_repr=r"\rm{a}", offset=0.0, prefixable=True)
        _ureg_unyt.add("j", 1.0, dimensions=unyt.dimensions.energy,
                       tex_repr=r"\rm{J}", offset=0.0, prefixable=True)
        # _ureg_unyt.add("cel", 1.0, dimensions=unyt.dimensions.temperature,
        #                tex_repr=r"^\circ\rm{C}", offset=-273.15, prefixable=True)
        # _ureg_unyt.add("j", 1.0, dimensions=unyt.dimensions.specific_flux,
        #                tex_repr=r"\rm{Jy}", prefixable=True)
        # _ureg_unyt.add("CH2O", 1.0, dimensions=unyt.dimensions.dimensionless,
        #                tex_repr=r"\rm{CH2O}", offset=0.0, prefixable=False)
        unyt._unit_lookup_table.inv_name_alternatives["acre"] = "ac"
        unyt._unit_lookup_table.inv_name_alternatives["are"] = "a"
        unyt._unit_lookup_table.inv_name_alternatives["hectare"] = "ha"
    return _ureg_unyt


def convert_to_pandas_timedelta(x):
    r"""Convert variable with time units to a pandas.Timedelta instance.

    Args:
        x (object): Scalar/array with units to convert to a pandas.Timedelta
            instance.

    Returns:
        pandas.Timedelta: Equivalent Timedelta variable.

    """
    assert(has_units(x))
    t_data = get_data(x)
    t_unit = get_units(x)
    unit_map = {'ns': 'ns',
                (tools.bytes2str(b'\xc2\xb5') + 's'): 'us',
                'ms': 'ms',
                's': 's',
                'min': 'm',
                'hr': 'h',
                'day': 'D'}
    return pd.Timedelta(t_data, unit=unit_map[t_unit])


def convert_from_pandas_timedelta(x):
    r"""Covert a pandas.Timedelta instance to a scalar/array with
    time units.

    Args:
        x (pandas.Timedelta): Timedelta variable to convert.

    Returns:
        object: Equivalent scalar/array with units.

    """
    return add_units(x.total_seconds(), 's')


def convert_matlab_unit_string(m_str):  # pragma: matlab
    r"""Convert Matlab unit string to string that the Python package
    can understand.

    Args:
        m_str (str): Matlab units string to convert.

    Returns:
        str: Converted string.

    """
    out = m_str
    replacements = {'h': 'hr'}
    regex_mu = tools.bytes2str(b'\xc2\xb5')
    regex = r'(?P<name>[A-Za-z%s]+)' % regex_mu
    for x in re.finditer(regex, m_str):
        xdict = x.groupdict()
        if xdict['name'] in replacements:
            xdict['name'] = replacements[xdict['name']]
            out = out[:(x.start())] + xdict['name'] + out[(x.end()):]
    return out


def convert_R_unit_string(r_str):
    r"""Convert R unit string to string that the Python package can
    understand.

    Args:
        r_str (str): R units string to convert.

    Returns:
        str: Converted string.

    """
    out = []
    replacements = {'h': 'hr'}
    regex_mu = tools.bytes2str(b'\xc2\xb5')
    regex = r'(?P<name>[A-Za-z%s]+)(?P<exp>-?[0-9]*)(?: |$)' % regex_mu
    for x in re.finditer(regex, r_str):
        xdict = x.groupdict()
        if xdict['name'] in replacements:
            xdict['name'] = replacements[xdict['name']]
        if xdict['exp']:
            out.append('({name}**{exp})'.format(**xdict))
        else:
            out.append(xdict['name'])
    return '*'.join(out)


def has_units(obj):
    r"""Determine if a Python object has associated units.

    Args:
        obj (object): Object to be tested for units.

    Returns:
        bool: True if the object has units, False otherwise.

    """
    out = hasattr(obj, 'units')
    if out and (obj.units == as_unit('dimensionless')):
        out = False
    return out


def get_units(obj):
    r"""Get the string representation of the units.

    Args:
        obj (object): Object to get units for.

    Returns:
        str: Units, empty if input object has none.
        
    """
    if has_units(obj):
        out = str(obj.units)
    else:
        out = ''
    return out


def get_data(obj):
    r"""Get the array/scalar assocaited with the object.

    Args:
        obj (object): Object to get data for.

    Returns:
        np.ndarray: Numpy array representation of the underlying data.

    """
    if has_units(obj):
        out = obj.to_ndarray()
        if out.ndim == 0:
            out = out.reshape((1, ))[0]
    else:
        out = obj
    return out


def add_units(arr, unit_str, dtype=None):
    r"""Add units to an array or scalar.

    Args:
        arr (np.ndarray, float, int): Scalar or array of data to add units to.
        unit_str (str): Unit string.
        dtype (np.dtype, optional): Numpy data type that should be maintained for
            array/qunatity with units. If not provided, this is determined from the
            array.

    Returns:
        unyt.unyt_array: Array with units.

    """
    ureg = get_ureg()
    unit_str = tools.bytes2str(unit_str)
    if is_null_unit(unit_str):
        return arr
    if has_units(arr):
        return convert_to(arr, unit_str)
    if dtype is None:
        if isinstance(arr, np.ndarray):
            dtype = arr.dtype
        else:
            dtype = np.array([arr]).dtype
    if isinstance(arr, np.ndarray) and (arr.ndim > 0):
        out = unyt.unyt_array(arr, unit_str, dtype=dtype,
                              registry=ureg)
    else:
        out = unyt.unyt_quantity(arr, unit_str, dtype=dtype,
                                 registry=ureg)
    return out


def are_compatible(units1, units2):
    r"""Check if two units are compatible.

    Args:
        units1 (str): First units string.
        units2 (str): Second units string.

    Returns:
        bool: True if the units are compatible, False otherwise.

    """
    # Empty units always compatible
    if is_null_unit(units1) or is_null_unit(units2):
        return True
    if (not is_unit(units1)) or (not is_unit(units2)):
        return False
    x = add_units(1, units1)
    try:
        convert_to(x, units2)
    except ValueError:
        return False
    return True


def is_null_unit(ustr):
    r"""Determines if a string is a null unit.

    Args:
        ustr (str): String to test.

    Returns:
        bool: True if the string is '' or 'n/a', False otherwise.

    """
    if (len(ustr) == 0) or (ustr == 'n/a'):
        return True
    return False


def as_unit(ustr):
    r"""Get unit object for the string.

    Args:
        ustr (str): Unit string.

    Returns:
        unyt.Unit: Unit object.

    Raises:
        ValueError: If the string is not a recognized unit.

    """
    try:
        out = unyt.Unit(ustr, registry=get_ureg())
    except unyt.exceptions.UnitParseError as e:
        raise ValueError(str(e))
    return out


def is_unit(ustr):
    r"""Determine if a string is a valid unit.

    Args:
        ustr (str): String representation to test.

    Returns:
        bool: True if the string is a valid unit. False otherwise.

    """
    ustr = tools.bytes2str(ustr)
    if is_null_unit(ustr):
        return True
    try:
        as_unit(ustr)
    except ValueError:
        return False
    return True


def convert_to(arr, new_units):
    r"""Convert quantity with units to new units. Objects without units
    will be returned with the new units.

    Args:
        arr (np.ndarray, float, int, unyt.unyt_array): Quantity with or
            without units.
        new_units (str): New units that should be applied.

    Returns:
        unyt.unyt_array: Array with new units.

    """
    if is_null_unit(new_units):
        return arr
    if not has_units(arr):
        return add_units(arr, new_units)
    try:
        out = arr.to(new_units)
    except unyt.exceptions.UnitConversionError as e:
        raise ValueError(str(e))
    return out


def get_conversion_function(old_units, new_units):
    r"""Get a function that will convert a scalar/array from one unit
    to another.

    Args:
        old_units (str): Units to convert from.
        new_units (str): Units to convert to.

    Returns:
        function: Conversion function that takes scalar/array as input
            and returns converted scalar/array.

    """
    def fconvert(x):
        ux = add_units(x, old_units)
        return get_data(convert_to(ux, new_units))
    return fconvert



import os
from yggdrasil.examples.tests import ExampleTstBase


class TestExampleSBML1(ExampleTstBase):
    r"""Base class for SBML example tests."""

    example_name = 'sbml1'

    @property
    def input_files(self):  # pragma: no cover
        r"""Input file."""
        return [os.path.join(self.yamldir, 'Input', 'input.txt')]

    @property
    def expected_output_files(self):
        r"""Expected output file."""
        return [os.path.join(self.yamldir, 'Output', 'expected.txt')]
        
    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.yamldir, 'output.txt')]
    

class TestExampleSBML2(TestExampleSBML1):
    r"""Test for SBML2 example."""
    
    example_name = 'sbml2'

    @property
    def expected_output_files(self):
        r"""Expected output file."""
        return [os.path.join(self.yamldir, 'Output', 'expected%d.txt')
                % i for i in range(2)]
        
    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.yamldir, 'output%d.txt')
                % i for i in range(2)]


class TestExampleSBML3(TestExampleSBML1):
    r"""Test for SBML3 example."""
    
    example_name = 'sbml3'



import os
import six
import uuid
import unittest
import tempfile
import shutil
import itertools
import flaky
from yggdrasil.components import ComponentMeta, import_component
from yggdrasil import runner, tools, platform
from yggdrasil.examples import yamls, source, ext_map
from yggdrasil.tests import YggTestBase, check_enabled_languages


_ext2lang = {v: k for k, v in ext_map.items()}
_test_registry = {}
_default_comm = tools.get_default_comm()


def iter_pattern_match(a, b):
    r"""Determine if two sets of iteration parameters match, allowing
    for wild cards.

    Args:
        a (tuple): Iteration parameters.
        b (tuple): Iteration parameters.

    Returns:
        bool: True if the parameters match, False otherwise.

    """
    assert(not isinstance(a, list))
    if isinstance(b, list):
        matches = [iter_pattern_match(a, ib) for ib in b]
        return any(matches)
    matches = []
    for ia, ib in zip(a, b):
        if not ((ia == ib) or (ia == '*') or (ib == '*')
                or (isinstance(ia, tuple) and (ib in ia))
                or (isinstance(ib, tuple) and (ia in ib))
                or (isinstance(ia, set) and (ib not in ia))
                or (isinstance(ib, set) and (ia not in ib))):
            return False
    return True


def make_iter_test(is_flaky=False, **kwargs):
    def itest(self):
        if is_flaky:
            self.sleep(1.0)
        self.run_iteration(**kwargs)
    if is_flaky:
        itest = flaky.flaky(max_runs=3)(itest)
    return itest


class ExampleMeta(ComponentMeta):

    def __new__(cls, name, bases, dct):
        iter_lists = []
        iter_keys = []
        test_name_fmt = 'test'
        iter_skip = dct.get('iter_skip', [])
        iter_flaky = dct.get('iter_flaky', [])
        iter_over = dct.get('iter_over', ['language'])
        iter_aliases = {'lang': 'language',
                        'type': 'datatype',
                        'types': 'datatype'}
        iter_over = [iter_aliases.get(x, x) for x in iter_over]
        for x in iter_over:
            test_name_fmt += '_%s'
            x_iter_list = dct.get('iter_list_%s' % x, None)
            for ibase in bases:
                if x_iter_list is not None:
                    break
                x_iter_list = getattr(ibase, 'iter_list_%s' % x, None)
            if x_iter_list is not None:
                iter_lists.append(x_iter_list)
                iter_keys.append(x)
            else:  # pragma: debug
                raise ValueError("Unsupported iter dimension: %s" % x)
        if dct.get('example_name', None) is not None:
            for x in itertools.product(*iter_lists):
                if iter_pattern_match(x, iter_skip):
                    continue
                itest_name = (test_name_fmt % x)
                if itest_name not in dct:
                    itest_func = make_iter_test(
                        is_flaky=iter_pattern_match(x, iter_flaky),
                        **{k: v for k, v in
                           zip(iter_keys, x)})
                    itest_func.__name__ = itest_name
                    dct[itest_name] = itest_func
        out = super(ExampleMeta, cls).__new__(cls, name, bases, dct)
        if out.example_name is not None:
            global _test_registry
            _test_registry[out.example_name] = out
        # else:
        #     out = unittest.skipIf(True, "Test uninitialized.")(out)
        return out


@six.add_metaclass(ExampleMeta)
class ExampleTstBase(YggTestBase, tools.YggClass):
    r"""Base class for running examples."""

    example_name = None
    expects_error = False
    env = {}
    iter_over = ['language']
    iter_skip = []
    iter_flaky = []
    iter_list_language = tools.get_supported_lang() + ['all', 'all_nomatlab']
    iter_list_comm = tools.get_supported_comm()
    iter_list_datatype = tools.get_supported_type()

    def __init__(self, *args, **kwargs):
        tools.YggClass.__init__(self, self.example_name)
        self.iter_param = {}
        self.uuid = str(uuid.uuid4())
        self.runner = None
        # self.debug_flag = True
        super(ExampleTstBase, self).__init__(*args, **kwargs)

    @property
    def language(self):
        r"""str: Language of the currect test."""
        return self.iter_param.get('language', None)

    @property
    def comm(self):
        r"""str: Comm used by the current test."""
        return self.iter_param.get('comm', _default_comm)

    @property
    def description_prefix(self):
        r"""Prefix message with test name."""
        return self.name

    @property
    def namespace(self):
        r"""str: Namespace for the example."""
        return "%s_%s" % (self.name, self.uuid)

    @property
    def tempdir(self):
        r"""str: Temporary directory."""
        return tempfile.gettempdir()

    @property
    def languages_tested(self):
        r"""list: Languages covered by the example."""
        if self.name not in source:  # pragma: debug
            return None
        if self.yaml is None:  # pragma: debug
            return None
        if self.language in ['all', 'all_nomatlab']:
            out = [_ext2lang[os.path.splitext(x)[-1]] for x in
                   source[self.name][self.language]]
        else:
            out = [self.language]
        return out

    @property
    def yaml(self):
        r"""str: The full path to the yaml file for this example."""
        if self.name not in yamls:  # pragma: debug
            return None
        if self.language in yamls[self.name]:
            return yamls[self.name][self.language]
        elif self.language.lower() in yamls[self.name]:
            return yamls[self.name][self.language.lower()]
        return None

    @property
    def yamldir(self):
        r"""str: Full path to the directory containing the yaml file."""
        if self.yaml is None:  # pragma: no cover
            return None
        if isinstance(self.yaml, list):
            out = os.path.dirname(self.yaml[0])
        else:
            out = os.path.dirname(self.yaml)
        return out

    # @property
    # def yaml_contents(self):
    #     r"""dict: Contents of yaml file."""
    #     if self.yaml is None:  # pragma: no cover
    #         return None
    #     return tools.parse_yaml(self.yaml)

    @property
    def input_files(self):  # pragma: debug
        r"""list: Input files for the run."""
        return None

    @property
    def expected_output_files(self):  # pragma: debug
        r"""list: Examples of expected output for the run."""
        return self.input_files

    @property
    def output_files(self):
        r"""list: Output files for the run."""
        return None

    @property
    def results(self):
        r"""list: Results that should be found in the output files."""
        if self.expected_output_files is None:  # pragma: debug
            return None
        out = []
        for fname in self.expected_output_files:
            assert(os.path.isfile(fname))
            out.append(self.read_file(fname))
        return out

    def check_results(self):
        r"""This should be overridden with checks for the result."""
        if self.output_files is None:
            return
        res_list = self.results
        out_list = self.output_files
        assert(res_list is not None)
        assert(out_list is not None)
        self.assert_equal(len(res_list), len(out_list))
        for res, fout in zip(res_list, out_list):
            self.check_file_exists(fout)
            if isinstance(res, tuple):
                res[0](fout, *res[1:])
            else:
                self.check_file_size(fout, res)
                self.check_file_contents(fout, res)

    def run_example(self):
        r"""This runs an example in the correct language."""
        if self.yaml is None:
            if self.name is not None:
                raise unittest.SkipTest("Could not locate example %s in language %s." %
                                        (self.name, self.language))
        else:
            # Check that language is installed
            for x in self.languages_tested:
                if not tools.is_lang_installed(x):
                    raise unittest.SkipTest("%s not installed." % x)
            # Copy platform specific makefile
            if self.language == 'make':
                makefile = os.path.join(self.yamldir, 'src', 'Makefile')
                if platform._is_win:  # pragma: windows
                    makedrv = import_component('model', 'make')
                    assert(makedrv.get_tool('compiler').toolname == 'nmake')
                    make_ext = '_windows'
                else:
                    make_ext = '_linux'
                shutil.copy(makefile + make_ext, makefile)
            # Check that comm is installed
            if self.comm in ['ipc', 'IPCComm']:
                from yggdrasil.communication.IPCComm import (
                    ipcrm_queues, ipc_queues)
                qlist = ipc_queues()
                if qlist:  # pragma: debug
                    print('Existing queues:', qlist)
                    ipcrm_queues()
            # Run
            os.environ.update(self.env)
            self.runner = runner.get_runner(self.yaml, namespace=self.namespace)
            self.runner.run()
            if self.expects_error:
                assert(self.runner.error_flag)
            else:
                assert(not self.runner.error_flag)
            try:
                self.check_results()
            finally:
                self.example_cleanup()
                # Remove copied makefile
                if self.language == 'make':
                    makefile = os.path.join(self.yamldir, 'src', 'Makefile')
                    if os.path.isfile(makefile):
                        os.remove(makefile)
                self.runner = None

    def example_cleanup(self):
        r"""Cleanup files created during the test."""
        if (self.yaml is not None) and (self.output_files is not None):
            timer_class = tools.YggClass()
            for fout in self.output_files:
                if os.path.isfile(fout):
                    tools.remove_path(fout, timer_class=timer_class, timeout=5)

    def setup_iteration(self, **kwargs):
        r"""Perform setup associated with an iteration."""
        for k, v in kwargs.items():
            k_setup = getattr(self, 'setup_iteration_%s' % k, None)
            if k_setup is not None:
                v = k_setup(v)
            self.iter_param[k] = v

    def teardown_iteration(self, **kwargs):
        r"""Perform teardown associated with an iteration."""
        for k, v in kwargs.items():
            k_teardown = getattr(self, 'teardown_iteration_%s' % k, None)
            if k_teardown is not None:
                k_teardown(v)
            del self.iter_param[k]
        assert(not self.iter_param)
        self.iter_param = {}

    def setup_iteration_language(self, language=None):
        r"""Perform setup associated with a language iteration."""
        if language is not None:
            check_enabled_languages(language)
        return language

    def setup_iteration_comm(self, comm=None):
        r"""Perform setup associated with a comm iteration."""
        assert(comm is not None)
        if not tools.is_comm_installed(comm):
            raise unittest.SkipTest("%s library not installed."
                                    % comm)
        self.set_default_comm(default_comm=comm)
        return comm

    def teardown_iteration_comm(self, comm=None):
        r"""Peform teardown associated with a comm iteration."""
        self.reset_default_comm()

    def run_iteration(self, **kwargs):
        r"""Run a test for the specified parameters."""
        if not tools.check_environ_bool('YGG_ENABLE_EXAMPLE_TESTS'):
            raise unittest.SkipTest("Example tests not enabled.")
        self.setup_iteration(**kwargs)
        try:
            getattr(self, kwargs.get('method', 'run_example'))()
        finally:
            self.teardown_iteration(**kwargs)



from yggdrasil.communication.transforms.TransformBase import TransformBase


class FunctionTransform(TransformBase):
    r"""Class for transforming messages based on a provided Python function.

    Args:
        function (func): The handle for a callable Python object (e.g. function)
            that should be used to transform messages or a string of the form
            "<function file>:<function name>" identifying a function where
            "<function file>" is the module or Python file containing the function
            and "<function name>" is the name of the function. The function should
            take the message as input and return the transformed message.

    """
    _transformtype = 'function'
    _schema_required = ['function']
    _schema_properties = {'function': {'type': 'function'}}

    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        return self.function(x)
    
    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """

        def ftran(x):
            if isinstance(x, list):
                return [ftran(ix) for ix in x]
            return x**3
        
        return [{'kwargs': {'function': ftran},
                 'in/out': [(1, 1), (2, 8)],
                 'in/out_t': [
                     ({'type': 'array',
                       'items': [
                           {'type': 'int', 'title': x,
                            'precision': 64, 'units': ''}
                           for x in 'abc']},
                      {'type': 'array',
                       'items': [
                           {'type': 'int', 'title': x,
                            'precision': 64, 'units': ''}
                           for x in 'abc']})]}]
